;; Generated package description from seq.el  -*- no-byte-compile: t -*-
(define-package "seq" "2.24.0.20240201.135317" "Sequence manipulation functions" 'nil :commit "27a90793a13f149121180e864fa53d68b9eac0b3" :url "https://elpa.gnu.org/packages/seq.html" :authors '(("Nicolas Petton" . "nicolas@petton.fr")) :maintainer '(nil . "emacs-devel@gnu.org") :keywords '("sequences"))
